#include "cache.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifdef WINDOZ
#include <io.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include "common.h"
#include "buffer.h"

static do_buf file_in;

static char *get_file_content(char *file)
{
  FILE *f = fopen(file, "rb");
  int c;
  char z;
  if (!f) return 0;
  reset_buffer(&file_in);
  while (1) {
    c = fgetc(f);
    if (c == EOF) break;
    z = c;
    add_buffer(&file_in, &z, 1);
  }
  fclose(f);
  z = 0;
  add_buffer(&file_in, &z, 1);
  return file_in.data;
}

/* n must be writable */
void create_dirs(char *n)
{
  /* if n = "a/b/c/d" create dirs a then a/b then a/b/c */
  /* no check is done, so pray God that all goes fine (or George Sodini,
   * as you prefer)
   */
  char *cur = n;
  while (1) {
    cur = strchr(cur, '/');
    if (!cur) break;
    *cur = 0;
#ifdef WINDOZ
    mkdir(n);
    *cur = '\\';
#else
    mkdir(n, 0755);
    *cur = '/';
#endif
    cur++;
  }
}

void save_file(char *name, char *data)
{
  FILE *f;
  create_dirs(name);
  f = fopen(name, "wb");
  if (!f) {
    perror(name);
    exit(1);
  }
  fprintf(f, "%s", data);
  fclose(f);
}

char *cache_get(cache_type t, int v1, int v2, int v3)
{
  char *r;
  char u[2048];
  char n[2048];
  switch (t) {
    case VIEWFORUM:
      sprintf(u, "/viewforum.php?f=%d&start=%d&sk=s", v1, v2); /* CHECKED */
      sprintf(n, "cache/f%d/v%d", v1, v2); /* CHECKED */
      break;
    case VIEWTOPIC:
      sprintf(u, "/viewtopic.php?f=%d&t=%d&start=%d", v1, v2, v3); /* CHECKED */
      sprintf(n, "cache/f%d/t%d/v%d", v1, v2, v3); /* CHECKED */
      break;
    case EDITPOST:
      sprintf(u, "/posting.php?mode=edit&f=%d&p=%d", v1, v3); /* CHECKED */
      sprintf(n, "cache/f%d/t%d/e%d", v1, v2, v3); /* CHECKED */
      break;
  }
  r = get_file_content(n);
  if (r) {
    printf("   [found in cache]\n");
    return r;
  }
again:
  r = get(u);
  if (!r) { printf("PANIC: weird error in cache_get\n"); exit(1); }
  if (strstr(r, "<body id=\"errorpage\">")) {
    printf("WARNING: the server returned an error page, let's try again.\n");
    goto again;
  }
  save_file(n, r);
  return r;
}
