#include "common.h"

#include <curl/curl.h>
#include <stdio.h>
#include <libxml/tree.h>

#include "buffer.h"
#include "xml.h"

static char *main_url;

static char *admin_login;
static char *admin_password;

static do_buf recdata;

char *forum_url(char *end)
{
  static char url[4096];
  if (strlen(end) + strlen(main_url) > 4095) {
    printf("PANIC: url too long\n");
    exit(1);
  }
  sprintf(url, "%s%s", main_url, end); /* CHECKED */
  return url;
}

char *forum_login_password(void)
{
  static char lp[2048];
  if (strlen(admin_login) + strlen(admin_password) +
      strlen("username=&password=&login=Login") > 2047) {
    printf("PANIC: login/password too long\n");
    exit(1);
  }
  sprintf(lp, "username=%s&password=%s&login=Login", /* CHECKED */
          admin_login, admin_password);
  return lp;
}

size_t get_data_cb(void *buffer, size_t size, size_t nmemb, void *userp)
{
  add_buffer(&recdata, buffer, size*nmemb);
  return size * nmemb;
}

void check_login(void)
{
  xmlDocPtr tree;
  xmlNodePtr root;

  add_buffer(&recdata, "", 1);

  tree = transform(recdata.data, "check_login.xslt");
  if (!tree) {
    printf("Error: could not login for some reason\n");
    printf("Did you provide the correct login/password?\n");
    printf("(Contact sed@free.fr for some help)\n");
    exit(1);
  }

  root = xmlDocGetRootElement(tree);

  if (!root->children) {
    printf("PANIC: not logged in, for some reason.\n");
printf("%s", recdata.data);
    exit(1);
  }

  xmlFreeDoc(tree);
}

char sid[128];

void login(void)
{
  char *p;
  CURLcode ret;
  int i;
  CURL *h = curl_easy_init();
  if (!h) {
    printf("curl_easy_init error\n");
    com_fini();
    exit(1);
  }

  printf("Logging in...\n");

  reset_buffer(&recdata);
  curl_easy_setopt(h, CURLOPT_URL, forum_url("/ucp.php?mode=login"));
  curl_easy_setopt(h, CURLOPT_COOKIEJAR, "cookies.txt");
  curl_easy_setopt(h, CURLOPT_WRITEFUNCTION, get_data_cb);
  curl_easy_setopt(h, CURLOPT_POSTFIELDS, forum_login_password());
/*
curl_easy_setopt(h, CURLOPT_VERBOSE, 1);
curl_easy_setopt(h, CURLOPT_HEADER, 1);
*/
  ret = curl_easy_perform(h);
  if (ret) {
    printf("error with curl_easy_perform\n");
  }
  curl_easy_cleanup(h);
  check_login();

  /* save sid for later use by acp_login */
  p = strstr(recdata.data, "sid=");
  if (!p) {
    printf("PANIC: no sid found\n");
    exit(1);
  }
  /* checking that p+4 has at least 32 valid bytes */
  /* add a null at the end */
  add_buffer(&recdata, "", 1);
  for (i=0; i<32; i++) if (p[i+4] == 0) {
    printf("PANIC: %s:%d:%s\n", __FILE__, __LINE__, __FUNCTION__);
    exit(1);
  }
  memcpy(sid, p+4, 32); /* CHECKED */
  sid[32] = 0;

  printf("...done\n");
}

void acp_login(void)
{
  char u[256];
  char v[512];
  char *p;
  char cred[64];
  xmlDocPtr tree;
  xmlNodePtr root;
  CURLcode ret;
  int i;
  CURL *h = curl_easy_init();
  if (!h) {
    printf("curl_easy_init error\n");
    com_fini();
    exit(1);
  }

  printf("Logging in in the ACP...\n");

  /* first step: get the credential */
  reset_buffer(&recdata);
  if (strlen(sid) + strlen("/adm/index.php?sid=") > 255) {
    printf("PANIC: string too long. %s:%d:%s\n",
           __FILE__, __LINE__, __FUNCTION__);
    exit(1);
  }
  sprintf(u, "/adm/index.php?sid=%s", sid); /* CHECKED */
  curl_easy_setopt(h, CURLOPT_URL, forum_url(u));
  curl_easy_setopt(h, CURLOPT_COOKIEFILE, "cookies.txt");
  curl_easy_setopt(h, CURLOPT_WRITEFUNCTION, get_data_cb);

  ret = curl_easy_perform(h);
  if (ret) {
    printf("error with curl_easy_perform\n");
    exit(1);
  }

  add_buffer(&recdata, "", 1);

  tree = transform(recdata.data, "login_first_step.xslt");
  if (!tree) {
ler:
    printf("Error: ACP login failed. Contact sed@free.fr for help.\n");
    exit(1);
  }

  root = xmlDocGetRootElement(tree);

  if (!root->children) { printf("fuck\n"); goto ler; }

  NP(root->children->children);
  NP(root->children->children->content);

  if (strlen(root->children->children->content) > 60) {
    printf("PANIC: %s:%d:%s\n", __FILE__, __LINE__, __FUNCTION__);
    exit(1);
  }
  strcpy(cred, root->children->children->content); /* CHECKED */
  xmlFreeDoc(tree);

  /* second step: do the login */
  reset_buffer(&recdata);
  /* redo the check? unnecessary? I won't think too much and do the check,
   * login is done only once, it's a non issue to do it twice
   */
  if (strlen(sid) + strlen("/adm/index.php?sid=") > 255) {
    printf("PANIC: string too long. %s:%d:%s\n",
           __FILE__, __LINE__, __FUNCTION__);
    exit(1);
  }
  sprintf(u, "/adm/index.php?sid=%s", sid); /* CHECKED */
  curl_easy_setopt(h, CURLOPT_URL, forum_url(u));
  if (strlen(admin_login) + strlen(cred) + strlen(admin_password)
     + strlen(cred) + strlen(sid)
     + strlen("username=&password_=&redirect=index.php&credential=&login=Login&sid=")
      > 511) {
    printf("PANIC: string too long. %s:%d:%s\n",
           __FILE__, __LINE__, __FUNCTION__);
    exit(1);
  }
  sprintf(v, /* CHECKED */
    "username=%s&password_%s=%s&redirect=index.php&credential=%s&login=Login&sid=%s",
    admin_login, cred, admin_password, cred, sid);
  curl_easy_setopt(h, CURLOPT_POSTFIELDS, v);
  curl_easy_setopt(h, CURLOPT_COOKIEJAR, "cookies.txt");

  ret = curl_easy_perform(h);
  if (ret) {
    printf("error 2 with curl_easy_perform\n");
    exit(1);
  }

  curl_easy_cleanup(h);

  check_login();

  /* save sid for later use */
  p = strstr(recdata.data, "sid=");
  if (!p) {
    printf("PANIC: no sid found\n");
    exit(1);
  }
  /* checking that p+4 has at least 32 valid bytes */
  /* add a null at the end */
  add_buffer(&recdata, "", 1);
  for (i=0; i<32; i++) if (p[i+4] == 0) {
    printf("PANIC: %s:%d:%s\n", __FILE__, __LINE__, __FUNCTION__);
    exit(1);
  }
  memcpy(sid, p+4, 32); /* CHECKED */
  sid[32] = 0;

  printf("sid is %s\n", sid);
  printf("...done\n");
}

void logout(void)
{
  CURLcode ret;
  CURL *h = curl_easy_init();
  if (!h) {
    printf("curl_easy_init error\n");
    com_fini();
    exit(1);
  }

  printf("Logging out...\n");

  reset_buffer(&recdata);
  curl_easy_setopt(h, CURLOPT_URL, forum_url("/ucp.php?mode=logout"));
  curl_easy_setopt(h, CURLOPT_COOKIEJAR, "cookies.txt");
  curl_easy_setopt(h, CURLOPT_WRITEFUNCTION, get_data_cb);
/*
curl_easy_setopt(h, CURLOPT_VERBOSE, 1);
curl_easy_setopt(h, CURLOPT_HEADER, 1);
*/
  ret = curl_easy_perform(h);
  if (ret) {
    printf("error with curl_easy_perform\n");
  }
  curl_easy_cleanup(h);

  printf("...done\n");
}

char *progname;
void usage(void)
{
  printf("Usage: %s [options]\n", progname);
  printf("\n");
  printf("Options:\n");
  printf("  --url <url>: give the base URL of the forum.\n");
  printf("               example: http://rcboat.ownforum.org\n");
  printf("  --login <login>: give, well, the login to get stuff from forum\n");
  printf("  --password <password>: guess what?\n");
  if (!strcmp(progname, "posts")) {
    printf("  --test: to test the program. Downloads at most 10 topics\n");
    printf("          and at most 10 posts per topic (so at most 100");
    printf(" posts).\n");
  }
  printf("\n");
  printf("Typcal use:\n");
  printf("  %s --url http://yoyo.com --login admin --password pass\n",
         progname);
  printf("\n");
  if (!strcmp(progname, "users")) {
    printf("The program will create users.txt and users.php\n");
    printf("Run \"php users.php > users.sql\" in this directory (very important)\n");
    printf("Put users.sql in store/ of your forum under a name that the\n");
    printf("restore will accept (do a backup and overwrite the file is a\n");
    printf("way to do it)\n");
  }
  exit(1);
}

int test_mode;
void com_init(char *progn, int n, char **v)
{
  int i=1;

  progname = progn;

  for (i=1; i<n; i++) {
    if (!strcmp(v[i], "--url")) {
      i++;
      if (i != n) main_url = v[i];
      continue;
    }
    if (!strcmp(v[i], "--login")) {
      i++;
      if (i != n) admin_login = v[i];
      continue;
    }
    if (!strcmp(v[i], "--password")) {
      i++;
      if (i != n) admin_password = v[i];
      continue;
    }
    if (!strcmp(v[i], "--test") && !strcmp(progn, "posts")) {
      extern int test_mode; test_mode = 1;
      continue;
    }
    if (!strcmp(v[i], "--help")) usage();
    usage();
  }

  if (!main_url) {
    printf("You must give the base URL of your forum.\n");
    exit(1);
  }
  if (!admin_login || !admin_password) {
    printf("You must give a login/password.\n");
    exit(1);
  }

  unlink("cookies.txt");

  if (curl_global_init(CURL_GLOBAL_ALL)) {
    printf("error init CURL");
    exit(1);
  }

  srand(time(0));
}

void com_fini(void)
{
  curl_global_cleanup();
}

/* return 0 if error */
char *get(char *url)
{
  CURLcode ret;
  CURL *h = curl_easy_init();
  if (!h) {
    printf("curl_easy_init error\n");
    com_fini();
    exit(1);
  }

  printf("Getting %s%s...\n", main_url, url);

  reset_buffer(&recdata);
  curl_easy_setopt(h, CURLOPT_URL, forum_url(url));
  curl_easy_setopt(h, CURLOPT_COOKIEFILE, "cookies.txt");
  curl_easy_setopt(h, CURLOPT_WRITEFUNCTION, get_data_cb);
  ret = curl_easy_perform(h);
  if (ret) {
    printf("error with curl_easy_perform\n");
    curl_easy_cleanup(h);
    return 0;
  }
  curl_easy_cleanup(h);
  add_buffer(&recdata, "", 1);

  printf("...done\n");

  return recdata.data;
}

char *get_sid(void)
{
  return sid;
}

do_buf sql_buf;
char *sql_escape(char *s)
{
  char t;
  reset_buffer(&sql_buf);
  while (*s) {
    switch (*s) {
#if 0
      case 0:
      case '\n':
      case '\r':
#endif
      case '\\':
      case '"':
#if 0
      case 0x1a:
#endif
        t = '\\';
        add_buffer(&sql_buf, &t, 1);
    }
    t = *s;
    add_buffer(&sql_buf, &t, 1);
    s++;
  }
  add_buffer(&sql_buf, "", 1);
  return sql_buf.data;
}

long make_date(char *day, char *month, char *year, char *hour, char *minute)
{
  time_t r;
  struct tm t;
  t.tm_sec = 0;
  sscanf(minute, "%d", &t.tm_min);
  sscanf(hour, "%d", &t.tm_hour);
  sscanf(day, "%d", &t.tm_mday);
  t.tm_mon = get_month(month);
  sscanf(year, "%d", &t.tm_year);
  t.tm_year -= 1900;
  t.tm_isdst = 0;
  r = mktime(&t);
  if (r == -1) r = 0;
  return r;
}

char *months[] = {
#include "months.h"
};

/* translates the greek month to english version
 * if not found, return as is
 */
char *translate_month(char *m)
{
  int cur = 0;
  while (months[cur]) {
    if (!strcmp(m, months[cur+1]))
      return months[cur];
    cur+=2;
  }
  return m;
}

int get_month(char *m)
{
  int i;
  for (i=0; months[i]; i+=2) {
    if (!strcmp(months[i], m)) {
      i /= 2;
      i %= 12;
      return i;
    }
  }
  printf("WARNING: unknown month '%s', pretend it's january\n", m);
  return 0;
}
