#include "common.h"

#include <stdlib.h>
#include <libxml/tree.h>
#include <string.h>
#include <time.h>

#include "xml.h"
#include "buffer.h"

typedef struct {
  char *name;
  char *desc;
  int type;
  int id;
  int parent_id; /* -1: no parent */
  int left_id;
  int right_id;
  int desc_parse_bbcode;
  int desc_parse_smilies;
  int desc_parse_urls;
  int status;
  int display_subforum_list;
  int display_on_index;
  int enable_post_review;	/* not used */
  int enable_indexing;
  int display_recent;		/* not used */
  int topics_per_page;
  int link_display_on_index;
  char *forum_link;		/* not used */
  int forum_link_track;
  char *forum_rules_link;
  char *forum_rules;
  int rules_parse_bbcode;
  int rules_parse_smilies;
  int rules_parse_urls;
} forum;

typedef struct {
  forum *data;
  int size;
  int maxsize;
} flist;

flist forums;

int max_right_id(void)
{
  int r = 0;
  int i;
  for (i=0; i<forums.size; i++)
    if (forums.data[i].right_id > r)
      r = forums.data[i].right_id;
  return r;
}

void left_right_id(int id, int *l, int *r)
{
  int i;
  for (i=0; i<forums.size; i++) {
    if (id == forums.data[i].id) {
      *l = forums.data[i].left_id;
      *r = forums.data[i].right_id;
      return;
    }
  }
  printf("PANIC: weird, in left_right_id\n");
  exit(1);
}

void add_forum(int id, int parent_id)
{
  if (forums.size == forums.maxsize) {
    forums.maxsize += 256;
    forums.data = realloc(forums.data, forums.maxsize * sizeof(forum));
    if (!forums.data) {
      printf("PANIC: out of memory\n");
      exit(1);
    }
  }

  memset(&forums.data[forums.size], 0, sizeof(forum));
  forums.data[forums.size].name = 0;
  forums.data[forums.size].id = id;
  forums.data[forums.size].parent_id = parent_id;

  /* left/right_id stuff, ripped from acp/acp_forums.php,
   * hopefully phpBB people won't change it...
   */
  if (parent_id == -1) {
    int id = max_right_id();
    forums.data[forums.size].left_id = id+1;
    forums.data[forums.size].right_id = id+2;
  } else {
    int i;
    int left_id, right_id;
    left_right_id(parent_id, &left_id, &right_id);
    /* update left/right ids */
    for (i=0; i<forums.size; i++) {
      if (forums.data[i].left_id > right_id) {
        forums.data[i].left_id += 2;
        forums.data[i].right_id += 2;
      }
    }
    for (i=0; i<forums.size; i++) {
      if (forums.data[i].left_id <= left_id &&
          left_id <= forums.data[i].right_id) {
        forums.data[i].right_id += 2;
      }
    }
    forums.data[forums.size].left_id = right_id;
    forums.data[forums.size].right_id = right_id+1;
  }

  forums.size++;
}

int exist_forum(int id)
{
  int i;
  for (i=0; i<forums.size; i++) if (forums.data[i].id == id) return 1;
  return 0;
}

void get_forums(int current_id, char *url)
{
  static char next_url[128];
  char *page;
  xmlDocPtr doc;
  xmlNodePtr root;
  xmlNodePtr cur;

  printf("Getting forum in: %s\n", url);

try_again:
  page = get(url);
  if (!page) exit(1);
  /* error pages are not to be parsed */
  if (strstr(page, "<body id=\"errorpage\">")) {
    printf("WARNING: the server returned an error page, let's try again.\n");
    goto try_again;
  }

//printf("%s", page); exit(1);

  /* extract forums */
  doc = transform(page, "get_forums_a.xslt");
  if (!doc) {
    printf("PANIC: %s:%d error with transform?\n", __FILE__, __LINE__);
    exit(1);
  }

  root = xmlDocGetRootElement(doc);
  if (!root) {
    printf("PANIC: no root element.\n");
    exit(1);
  }

  cur = root->children;
  if (!cur) {
    printf("No child forum for %s\n", forum_url(url));
    goto end;
  }

  while(cur) {
    int id;
    NP(cur->children->children);
    NP(cur->children->children->content);
    if (strlen(cur->children->children->content) + strlen("/viewforum.php?f=")
         > 127) {
      printf("PANIC: string too long. %s:%d:%s\n",
             __FILE__, __LINE__, __FUNCTION__);
      exit(1);
    }
    sprintf(next_url, /* CHECKED */
            "/viewforum.php?f=%s", cur->children->children->content);
    sscanf(cur->children->children->content, "%d", &id);
    if (!exist_forum(id)) {
      add_forum(id, current_id);
      get_forums(id, next_url);
    }
    cur = cur->next;
  }

end:
  xmlFreeDoc(doc);
}

void get_forums_info(void)
{
  xmlDocPtr doc;
  xmlNodePtr root;
  xmlNodePtr cur;
  char *d;
  char v[512];

  int i;

  acp_login();

  for (i = 0; i<forums.size; i++) {
    /* should check... */
    if (20 + 20 + strlen(get_sid())
 + strlen("/adm/index.php?i=forums&mode=manage&parent_id=&f=&action=edit&sid=")
         > 511) {
      printf("PANIC: string too long. %s:%d:%s\n",
             __FILE__, __LINE__, __FUNCTION__);
      exit(1);
    }
    sprintf(v, /* CHECKED */
    "/adm/index.php?i=forums&mode=manage&parent_id=%d&f=%d&action=edit&sid=%s",
      forums.data[i].parent_id != -1 ? forums.data[i].parent_id : 0,
      forums.data[i].id,
      get_sid());
try_again:
    d = get(v);
    /* error pages are not to be parsed */
    if (strstr(d, "<body id=\"errorpage\">")) {
      printf("WARNING: the server returned an error page, let's try again.\n");
      goto try_again;
    }

    /* get data */
    printf("  ...getting data\n");
    doc = transform(d, "get_forum.xslt");
    if (!doc) {
      printf("%s\n", d);
      printf("PANIC: %s:%d error with transform?\n", __FILE__, __LINE__);
      exit(1);
    }
    root = xmlDocGetRootElement(doc);
    if (!root) {
      printf("%s\n", d);
      printf("PANIC: no root element.\n");
      exit(1);
    }
    cur = root->children;
    if (!cur) {
      printf("%s\n", d);
      printf("PANIC: bad forum file? (contact sed@free.fr)\n");
      exit(1);
    }
    forums.data[i].name = "";
    forums.data[i].desc = "";
    forums.data[i].forum_link = "";
    forums.data[i].forum_rules_link = "";
    forums.data[i].forum_rules = "";
    /* some pointers are not NP checked for they may be NULL
     * (but they are handled in if(cur->....->chidren->xxx)
     * so no crash possible)
     */
    while (cur) {
      NP(cur->name);
      if (!strcmp(cur->name, "type")) {
        printf("    ...found type\n");
        NP(cur->children);
        NP(cur->children->content);
        sscanf(cur->children->content, "%d", &forums.data[i].type);
        printf("      type=%d\n", forums.data[i].type);
      } else if (!strcmp(cur->name, "name")) {
        if (cur->children && cur->children->content) {
          printf("    ...found name\n");
          forums.data[i].name = xmlNodeListGetRawString(doc, cur->children, 0);
          if (!forums.data[i].name) {
            printf("PANIC: out of memory\n");
            exit(1);
          }
          printf("      name=%s\n", forums.data[i].name);
        }
      } else if (!strcmp(cur->name, "desc")) {
        if (cur->children && cur->children->content) {
          printf("    ...found desc\n");
          forums.data[i].desc = xmlNodeListGetRawString(doc, cur->children, 0);
          if (!forums.data[i].desc) {
            printf("PANIC: out of memory\n");
            exit(1);
          }
          printf("      desc=%s\n", forums.data[i].desc);
        }
      } else if (!strcmp(cur->name, "desc_parse_bbcode")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found desc_parse_bbcode\n");
        sscanf(cur->children->content, "%d", &forums.data[i].desc_parse_bbcode);
        printf("      desc_parse_bbcode=%d\n", forums.data[i].desc_parse_bbcode);
      } else if (!strcmp(cur->name, "desc_parse_smilies")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found desc_parse_smilies\n");
        sscanf(cur->children->content, "%d", &forums.data[i].desc_parse_smilies);
        printf("      desc_parse_smilies=%d\n", forums.data[i].desc_parse_smilies);
      } else if (!strcmp(cur->name, "desc_parse_urls")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found desc_parse_urls\n");
        sscanf(cur->children->content, "%d", &forums.data[i].desc_parse_urls);
        printf("      desc_parse_urls=%d\n", forums.data[i].desc_parse_urls);
      } else if (!strcmp(cur->name, "status")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found status\n");
        sscanf(cur->children->content, "%d", &forums.data[i].status);
        printf("      status=%d\n", forums.data[i].status);
      } else if (!strcmp(cur->name, "display_subforum_list")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found display_subforum_list\n");
        sscanf(cur->children->content, "%d", &forums.data[i].display_subforum_list);
        printf("      display_subforum_list=%d\n", forums.data[i].display_subforum_list);
      } else if (!strcmp(cur->name, "display_on_index")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found display_on_index\n");
        sscanf(cur->children->content, "%d", &forums.data[i].display_on_index);
        printf("      display_on_index=%d\n", forums.data[i].display_on_index);
      } else if (!strcmp(cur->name, "enable_post_review")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found enable_post_review\n");
        sscanf(cur->children->content, "%d", &forums.data[i].enable_post_review);
        printf("      enable_post_review=%d\n", forums.data[i].enable_post_review);
      } else if (!strcmp(cur->name, "enable_indexing")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found enable_indexing\n");
        sscanf(cur->children->content, "%d", &forums.data[i].enable_indexing);
        printf("      enable_indexing=%d\n", forums.data[i].enable_indexing);
      } else if (!strcmp(cur->name, "display_recent")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found display_recent\n");
        sscanf(cur->children->content, "%d", &forums.data[i].display_recent);
        printf("      display_recent=%d\n", forums.data[i].display_recent);
      } else if (!strcmp(cur->name, "topics_per_page")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found topics_per_page\n");
        sscanf(cur->children->content, "%d", &forums.data[i].topics_per_page);
        printf("      topics_per_page=%d\n", forums.data[i].topics_per_page);
      } else if (!strcmp(cur->name, "link_display_on_index")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found link_display_on_index\n");
        sscanf(cur->children->content, "%d", &forums.data[i].link_display_on_index);
        printf("      link_display_on_index=%d\n", forums.data[i].link_display_on_index);
      } else if (!strcmp(cur->name, "forum_link")) {
        if (cur->children && cur->children->content) {
          printf("    ...found forum_link\n");
          forums.data[i].forum_link = xmlNodeListGetRawString(doc, cur->children, 0);
          if (!forums.data[i].forum_link) {
            printf("PANIC: out of memory\n");
            exit(1);
          }
          printf("      forum_link=%s\n", forums.data[i].forum_link);
        }
      } else if (!strcmp(cur->name, "forum_link_track")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found forum_link_track\n");
        sscanf(cur->children->content, "%d", &forums.data[i].forum_link_track);
        printf("      forum_link_track=%d\n", forums.data[i].forum_link_track);
      } else if (!strcmp(cur->name, "forum_rules_link")) {
        if (cur->children && cur->children->content) {
          printf("    ...found forum_rules_link\n");
          forums.data[i].forum_rules_link = xmlNodeListGetRawString(doc, cur->children, 0);
          if (!forums.data[i].forum_rules_link) {
            printf("PANIC: out of memory\n");
            exit(1);
          }
          printf("      forum_rules_link=%s\n", forums.data[i].forum_rules_link);
        }
      } else if (!strcmp(cur->name, "forum_rules")) {
        if (cur->children && cur->children->content) {
          printf("    ...found forum_rules\n");
          forums.data[i].forum_rules = xmlNodeListGetRawString(doc, cur->children, 0);
          if (!forums.data[i].forum_rules) {
            printf("PANIC: out of memory\n");
            exit(1);
          }
          printf("      forum_rules=%s\n", forums.data[i].forum_rules);
        }
      } else if (!strcmp(cur->name, "rules_parse_bbcode")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found rules_parse_bbcode\n");
        sscanf(cur->children->content, "%d", &forums.data[i].rules_parse_bbcode);
        printf("      rules_parse_bbcode=%d\n", forums.data[i].rules_parse_bbcode);
      } else if (!strcmp(cur->name, "rules_parse_smilies")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found rules_parse_smilies\n");
        sscanf(cur->children->content, "%d", &forums.data[i].rules_parse_smilies);
        printf("      rules_parse_smilies=%d\n", forums.data[i].rules_parse_smilies);
      } else if (!strcmp(cur->name, "rules_parse_urls")) {
        NP(cur->children);
        NP(cur->children->content);
        printf("    ...found rules_parse_urls\n");
        sscanf(cur->children->content, "%d", &forums.data[i].rules_parse_urls);
        printf("      rules_parse_urls=%d\n", forums.data[i].rules_parse_urls);
      }
      cur = cur->next;
    }

    xmlFreeDoc(doc);
  }
}

#define OPTION_FLAG_BBCODE	1
#define OPTION_FLAG_SMILIES	2
#define OPTION_FLAG_LINKS	4

void save_forums(char *filename)
{
  FILE *f = fopen("forums.php", "wb");
  FILE *g = fopen("forums.txt", "wb");

  int i;

  fprintf(f, "<?php\n");

  for (i=0; i<forums.size; i++) {
    fprintf(g, "%d %d\n", forums.data[i].id, forums.data[i].parent_id);
    fprintf(f, "$m = new parse_message();\n");
    fprintf(f, "$m->message=\"%s\";\n", sql_escape(forums.data[i].desc));
    fprintf(f, "$m->parse(%s, %s, %s, true, true, true, true);\n",
            forums.data[i].desc_parse_bbcode ? "true" : "false",
            forums.data[i].desc_parse_urls ? "true" : "false",
            forums.data[i].desc_parse_smilies ? "true" : "false");
    fprintf(f, "fwrite($f, \"('\");\n");
    fprintf(f, "fwrite($f, mysql_real_escape_string(\"%s\"));\n",
            sql_escape(forums.data[i].name));
    fprintf(f, "fwrite($f, \"', '\");\n");
    fprintf(f, "fwrite($f, mysql_real_escape_string($m->message));\n");
    fprintf(f, "fwrite($f, \"', \");\n");
    fprintf(f, "fwrite($f, \"%d, %d, %d, %d, %d, '', '\");\n",
            forums.data[i].type,
            forums.data[i].id,
            forums.data[i].parent_id == -1 ? 0 : forums.data[i].parent_id,
            forums.data[i].left_id,
            forums.data[i].right_id);
    fprintf(f, "fwrite($f, $m->bbcode_uid);\n");
    fprintf(f, "fwrite($f, \"', '\");\n");
    fprintf(f, "fwrite($f, $m->bbcode_bitfield);\n");
    fprintf(f, "fwrite($f, \"'\");\n");
    fprintf(f, "fwrite($f, \", %d\");\n",
            (forums.data[i].desc_parse_bbcode ? OPTION_FLAG_BBCODE : 0) +
            (forums.data[i].desc_parse_smilies ? OPTION_FLAG_SMILIES : 0) +
            (forums.data[i].desc_parse_urls ? OPTION_FLAG_LINKS : 0));
    fprintf(f, "fwrite($f, \", %d\");\n", forums.data[i].status);
    fprintf(f, "fwrite($f, \", %d\");\n", forums.data[i].display_subforum_list);
    /* if forum = FORUM_LINK (which is = 2) then use link_display_on_index */
    fprintf(f, "fwrite($f, \", %d\");\n",
            forums.data[i].type == 2 ?
              forums.data[i].link_display_on_index :
              forums.data[i].display_on_index);
    fprintf(f, "fwrite($f, \", %d\");\n", forums.data[i].enable_indexing);
    fprintf(f, "fwrite($f, \", %d\");\n", forums.data[i].topics_per_page);
    fprintf(f, "fwrite($f, \", '\");\n");
    fprintf(f, "fwrite($f, mysql_real_escape_string(\"%s\"));\n",
            sql_escape(forums.data[i].forum_link));
    fprintf(f, "fwrite($f, \"'\");\n");
    fprintf(f, "fwrite($f, \", '\");\n");
    fprintf(f, "fwrite($f, mysql_real_escape_string(\"%s\"));\n",
            sql_escape(forums.data[i].forum_rules_link));
    fprintf(f, "fwrite($f, \"'\");\n");
    fprintf(f, "$m = new parse_message();\n");
    fprintf(f, "$m->message=\"%s\";\n", sql_escape(forums.data[i].forum_rules));
    fprintf(f, "$m->parse(%s, %s, %s, true, true, true, true);\n",
            forums.data[i].rules_parse_bbcode ? "true" : "false",
            forums.data[i].rules_parse_urls ? "true" : "false",
            forums.data[i].rules_parse_smilies ? "true" : "false");
    fprintf(f, "fwrite($f, \", '\");\n");
    fprintf(f, "fwrite($f, mysql_real_escape_string($m->message));\n");
    fprintf(f, "fwrite($f, \"', '\");\n");
    fprintf(f, "fwrite($f, $m->bbcode_uid);\n");
    fprintf(f, "fwrite($f, \"', '\");\n");
    fprintf(f, "fwrite($f, $m->bbcode_bitfield);\n");
    fprintf(f, "fwrite($f, \"'\");\n");
    fprintf(f, "fwrite($f, \", %d\");\n",
            (forums.data[i].rules_parse_bbcode ? OPTION_FLAG_BBCODE : 0) +
            (forums.data[i].rules_parse_smilies ? OPTION_FLAG_SMILIES : 0) +
            (forums.data[i].rules_parse_urls ? OPTION_FLAG_LINKS : 0));
    printf("FORUM %d\n", i);
    printf(
     "  name=%s\n  desc=%s\n  type=%d\n  id=%d\n  parent_id=%d\n  l,r=%d %d\n",
      forums.data[i].name,
      forums.data[i].desc,
      forums.data[i].type,
      forums.data[i].id,
      forums.data[i].parent_id,
      forums.data[i].left_id,
      forums.data[i].right_id);
    fprintf(f, "fwrite($f, \")%c\\n\");\n", i==forums.size-1 ? ';' : ',');
  }

  fprintf(f, "?>\n");
  fclose(f);
  fclose(g);
}

#include <stdio.h>

int main(int n, char **v)
{
  char *out = "forum.sql";

  com_init("forums", n, v);

  login();
  get_forums(-1, "/index.php");
  get_forums_info();
  save_forums(out);
  logout();

  com_fini();

  return 0;
}
