#include "common.h"

#include <stdlib.h>
#include <libxml/tree.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#include "xml.h"
#include "buffer.h"
#include "cache.h"

#include <stdio.h>

/* set to 1 in common.c if --test is passed, so that we do
 * only the 10 first topics and 10 first posts
 * (guess what? it's a test mode!)
 */
extern int test_mode;

/* for test_mode counting */
int nb_topics;
int nb_posts_per_topic;

static do_buf users_buf;

int find_user(char *u)
{
  char *b = users_buf.data;
  char *n;
  int i;
  int l = strlen(u);
more:
  if (!(b=strstr(b, u))) {
    printf("WARNING: users %s not found, pretending Anonymous\n", u);
    return 1;
  }
  /* we have to be sure it's the right name */
  /* we must have a space, a number, a 8char password, a space */
  /* (it's not enough, but I doubt we'll get a weird user's name
   * containing such a pattern...)
   */
  n = b + l;
  if (*n != ' ') goto try_next;
  n++;
  if (!isdigit(*n)) goto try_next;
  while (isdigit(*n)) n++;
  if (*n != ' ') goto try_next;
  n++;
  for (i=0; i<8; i++) if (n[i] != '_' && (n[i] < 'a' || n[i] > 'z') &&
                          (n[i] < '0' || n[i] > '9')) goto try_next;
  if (n[8] != ' ') goto try_next;
  /* we found it */
  sscanf(b+l, "%d", &i);
  return i;
try_next:
  b++;
  goto more;
}

void read_users(void)
{
  int c;
  char z;
  FILE *f = fopen("users.txt", "rb");
  if (!f) {
    perror("users.txt");
    exit(1);
  }
  while (1) {
    c = fgetc(f);
    if (c==EOF) break;
    z = c;
    add_buffer(&users_buf, &z, 1);
  }
  z = 0;
  add_buffer(&users_buf, &z, 1);
}

typedef struct {
  int topic_id;
  int forum_id;
  int post_id;
  int poster;
  long time;
  int is_master;	/* if 1 then title of post is title of topic */
  int icon;
  char *subject;
  char *content;
  int do_bbcode;
  int do_smilies;
  int do_magic_url;
  int do_attach_sig;
  int do_lock_topic;
  int do_lock_post;
  int type;
  int stick_topic_time_limit;
  char *poster_name;
  char *poster_colour;
} post;

typedef struct {
  int topic_id;
  int forum_id;
  char *topic_title;
  int topic_poster;
  int topic_type;
  /* populated later */
  long topic_time;
  int topic_locked;
  int icon_id;
  int has_poll;		/* 0: no, 1: yes */
  char *poll_question;
  char *poll_options;
  int options_per_user;
  int poll_length;
  int poll_vote_change;
  /* data computed on the fly */
  int topic_replies; /* topic_replies_real is same? */
  int topic_first_post_id;
  char *topic_first_poster_name;
  char *topic_first_poster_colour;
  int topic_last_post_id;
  int topic_last_poster_id;
  char *topic_last_poster_name;
  char *topic_last_poster_colour;
  char *topic_last_post_subject;
  long topic_last_post_time;
} topic;

post *pl;
int pl_size;
int pl_maxsize;

topic *tl;
int tl_size;
int tl_maxsize;

/* this was added after populate_topic_from_post
 * yeah, it's weird to have too functions like that
 * but this one is for counting. At first I did not
 * count anything, thinking phpbb had a "reset stats"
 * function to do that, but no... so here is some code
 * (and fuck! it's a hack! not a corporate crap, live
 * with that or do like George Sodini)
 */
void update_topic(topic *t, post *p)
{
  if (!p->is_master) {
    t->topic_replies++;
  } else {
    t->topic_first_post_id = p->post_id;
    t->topic_first_poster_name = p->poster_name;
    t->topic_first_poster_colour = p->poster_colour;
  }
  t->topic_last_post_id = p->post_id;
  t->topic_last_poster_id = p->poster;
  t->topic_last_poster_name = p->poster_name;
  t->topic_last_poster_colour = p->poster_colour;
  t->topic_last_post_subject = p->subject;
  t->topic_last_post_time = p->time;
}

topic *find_topic(int forum_id, int topic_id)
{
  int i;

  /* only checking topic_id */
  for (i=0; i<tl_size; i++)
//    if (tl[i].topic_id == topic_id && tl[i].forum_id == forum_id)
    if (tl[i].topic_id == topic_id)
      return tl+i;

  printf("FATAL: %s\n", __FUNCTION__);
  exit(1);
}

void add_post(int forum_id, int topic_id, int post_id, int user_id, long time,
              int is_master, char *poster_name, char *poster_color)
{
  topic *t = find_topic(forum_id, topic_id);
  int i;
  /* is post already here? */
  /* only check post_id */
  for (i=0; i<pl_size; i++) {
    //if (pl[i].topic_id == topic_id && pl[i].forum_id == forum_id &&
    //    pl[i].post_id == post_id) {
    if (pl[i].post_id == post_id) {
      printf("warning: post %d (from forum %d, topic %d) already seen\n",
             post_id, pl[i].forum_id, pl[i].topic_id);
      return;
    }
  }
  /* add it */
  if (pl_size == pl_maxsize) {
    pl_maxsize += 32768;
    pl = realloc(pl, pl_maxsize * sizeof(post));
    if (!pl) { printf("PANIC: out of memory\n"); exit(1); }
  }
  pl[pl_size].topic_id = topic_id;
  pl[pl_size].forum_id = forum_id;
  pl[pl_size].post_id = post_id;
  pl[pl_size].poster = user_id;
  pl[pl_size].time = time;
  pl[pl_size].is_master = is_master;
  pl[pl_size].icon = 0;
  pl[pl_size].subject = "";
  pl[pl_size].content = "";
  pl[pl_size].do_bbcode = 1;
  pl[pl_size].do_smilies = 1;
  pl[pl_size].do_magic_url = 1;
  pl[pl_size].do_attach_sig = 1;
  pl[pl_size].do_lock_topic = 0;
  pl[pl_size].do_lock_post = 0;
  pl[pl_size].type = 0;
  pl[pl_size].poster_name = poster_name;
  pl[pl_size].poster_colour = poster_color;
  /* color might be null in some cases */
  if (!pl[pl_size].poster_colour) pl[pl_size].poster_colour = "000000";

  update_topic(t, pl+pl_size);

  pl_size++;
}

void add_topic(int forum_id, int topic_id, int user_id)
{
  int i;
  /* is topic already here? */
  /* check only topic_id */
  for (i=0; i<tl_size; i++) {
    //if (tl[i].topic_id == topic_id && tl[i].forum_id == forum_id) {
    if (tl[i].topic_id == topic_id) {
      printf("warning: topic %d (from forum %d) already seen\n",
             topic_id, tl[i].forum_id);
      return;
    }
  }
  /* add it */
  if (tl_size == tl_maxsize) {
    tl_maxsize += 1024;
    tl = realloc(tl, tl_maxsize * sizeof(topic));
    if (!tl) { printf("PANIC: out of memory\n"); exit(1); }
  }
  tl[tl_size].topic_id = topic_id;
  tl[tl_size].forum_id = forum_id;
  tl[tl_size].topic_title = "";
  tl[tl_size].topic_poster = user_id;
  tl[tl_size].topic_time = 0;
  tl[tl_size].topic_type = 0;
  tl[tl_size].has_poll = 0;
  tl[tl_size].topic_replies = 0;
  tl[tl_size].topic_first_post_id = 0;
  tl[tl_size].topic_first_poster_name = "";
  tl[tl_size].topic_first_poster_colour = "000000";
  tl[tl_size].topic_last_post_id = 0;
  tl[tl_size].topic_last_poster_id = 0;
  tl[tl_size].topic_last_poster_name = "";
  tl[tl_size].topic_last_poster_colour = "000000";
  tl[tl_size].topic_last_post_subject = "";
  tl[tl_size].topic_last_post_time = 0;
  tl_size++;
}

typedef struct {
  int forum_id;
  int parent_id;	/* -1 if no parent */
  int depth;		/* -1 at start */
  int forum_posts;
  int forum_topics;	/* forum_topic_reals == forum_topics? */
  int forum_last_post_id;
  int forum_last_poster_id;
  char *forum_last_post_subject;
  long forum_last_post_time;
  char *forum_last_poster_name;
  char *forum_last_poster_colour;
} forum;

forum *forum_list;
int fl_size;
int fl_maxsize;

void add_forum(int forum_id, int parent_id)
{
  if (fl_size == fl_maxsize) {
    fl_maxsize += 256;
    forum_list = realloc(forum_list, fl_maxsize * sizeof(forum));
    if (!forum_list) { printf("out of memory\n"); exit(1); }
  }
  forum_list[fl_size].forum_id = forum_id;
  forum_list[fl_size].parent_id = parent_id;
  forum_list[fl_size].depth = -1;
  forum_list[fl_size].forum_posts = 0;
  forum_list[fl_size].forum_topics = 0;
  forum_list[fl_size].forum_last_post_id = 0;
  forum_list[fl_size].forum_last_poster_id = 0;
  forum_list[fl_size].forum_last_post_subject = "";
  forum_list[fl_size].forum_last_post_time = 0;
  forum_list[fl_size].forum_last_poster_name = "";
  forum_list[fl_size].forum_last_poster_colour = "000000";
  fl_size++;
}

void get_topics(int f)
{
  int start = 0;
  int nstart;
  char *d;
  xmlDocPtr doc;
  xmlNodePtr root;
  xmlNodePtr cur;

more:
  if (test_mode && nb_topics == 10) return;

  printf("Getting topics of forum %d (start=%d)\n", f, start);
  d = cache_get(VIEWFORUM, f, start, 0);

  /* get topics in this page */
  doc = transform(d, "posts_get_topics.xslt");
  if (!doc) {
    printf("PANIC: %s:%d error with transform?\n", __FILE__, __LINE__);
    exit(1);
  }
  root = xmlDocGetRootElement(doc);
  if (!root) {
    printf("PANIC: no root element.\n");
    exit(1);
  }
  cur = root->children;
  if (!cur)
    printf("WARNING: no topics for forum %d, verify that if it is normal!\n",
           f);
  while (cur && !(test_mode && nb_topics == 10)) {
    char *user;
//    char *color = xmlNodeListGetRawString(doc, cur->children->next->next->next->next->children, 0);
    int forum_id;
    int topic_id;
    int user_id;
    NP(cur->children);
    NP(cur->children->children);
    NP(cur->children->children->content);
    NP(cur->children->next);
    NP(cur->children->next->children);
    NP(cur->children->next->children->content);
    sscanf(cur->children->children->content, "%d", &topic_id);
    sscanf(cur->children->next->children->content, "%d", &forum_id);
    NP(cur->children->next->next);
    if (!cur->children->next->next->children) {
      printf(
  "WARNING: user not set! Let's pretend Anonymous... (topic %d)\n", topic_id);
      user = "Anonymous";
    } else
      user = xmlNodeListGetRawString(doc, cur->children->next->next->children, 0);
    user_id = find_user(user);
    printf("Adding topic %d (f=%d, u=%d) (user=%s)\n",
           topic_id, forum_id, user_id, user);
    free(user);
    add_topic(forum_id, topic_id, user_id);
    if (test_mode) nb_topics++;
    cur = cur->next;
  }
  xmlFreeDoc(doc);

  /* go next page (if any) */
  doc = transform(d, "posts_next_topics.xslt");
  if (!doc) {
    printf("PANIC: %s:%d error with transform?\n", __FILE__, __LINE__);
    exit(1);
  }
  root = xmlDocGetRootElement(doc);
  if (!root) {
    printf("PANIC: no root element.\n");
    exit(1);
  }
  cur = root->children;
  if (cur && cur->children && cur->children->content) {
    nstart = 0;
    sscanf(cur->children->content, "%d", &nstart);
    if (nstart > start) {
      start = nstart;
      xmlFreeDoc(doc);
      goto more;
    }
  }
  xmlFreeDoc(doc);
}

void get_forums(void)
{
  FILE *f = fopen("forums.txt", "rb");
  int n, p;

  printf("Getting list of forums...\n");

  /* first - read the list of forums */
  if (!f) {
    printf("error with forums.txt, impossible to open\n");
    exit(1);
  }
  while (fscanf(f, "%d %d\n", &n, &p) == 2) add_forum(n, p);
  fclose(f);

  printf("...done\n");

  /* second - proceed, one by one */
  for (n=0; n<fl_size; n++)
    get_topics(forum_list[n].forum_id);
}

void get_posts(topic *t)
{
  int start = 0;
  int first_post = 1;
  int nstart;
  char *d;
  xmlDocPtr doc;
  xmlNodePtr root;
  xmlNodePtr cur;

more:
  if (test_mode && nb_posts_per_topic == 10) return;

  printf("Getting posts of topic %d (start=%d)\n", t->topic_id, start);
  d = cache_get(VIEWTOPIC, t->forum_id, t->topic_id, start);

  /* get posts in this page */
  doc = transform(d, "posts_get_posts.xslt");
  if (!doc) {
    printf("PANIC: %s:%d error with transform?\n", __FILE__, __LINE__);
    exit(1);
  }
  root = xmlDocGetRootElement(doc);
  if (!root) {
    printf("PANIC: no root element.\n");
    exit(1);
  }
  cur = root->children;
  NP(cur);
  while (cur && !(test_mode && nb_posts_per_topic == 10)) {
    char *user;
    char *color;
    int forum_id = t->forum_id;
    int topic_id = t->topic_id;
    int post_id;
    int user_id;
    xmlNodePtr dt;
    char *day;
    char *month;
    char *year;
    char *hour;
    char *min;
    NP(cur->children);
    NP(cur->children->next);
    NP(cur->children->next->next);
    NP(cur->children->next->next->children);
    dt = cur->children->next->next->children;
    NP(dt->children);
    NP(dt->children->content);
    day = dt->children->content;
    NP(dt->next);
    NP(dt->next->children);
    NP(dt->next->children->content);
    month = dt->next->children->content;
    NP(dt->next->next);
    NP(dt->next->next->children);
    NP(dt->next->next->children->content);
    year = dt->next->next->children->content;
    NP(dt->next->next->next);
    NP(dt->next->next->next->children);
    NP(dt->next->next->next->children->content);
    hour = dt->next->next->next->children->content;
    NP(dt->next->next->next->next);
    NP(dt->next->next->next->next->children);
    NP(dt->next->next->next->next->children->content);
    min = dt->next->next->next->next->children->content;
    long time = make_date(day, month, year, hour, min);
    NP(cur->children->children);
    NP(cur->children->children->content);
    sscanf(cur->children->children->content, "%d", &post_id);
    if (!cur->children->next->children) {
      printf(
  "WARNING: user not set! Let's pretend Anonymous... (post id %d)\n", post_id);
      user = "Anonymous";
    } else
      user = xmlNodeListGetRawString(doc, cur->children->next->children, 0);
    NP(cur->children->next->next->next);
    if (!cur->children->next->next->next->children) {
      printf(
  "WARNING: user color not found! Let's pretend black... (post id %d)\n", post_id);
      /* passing 0 will result in "000000" for color */
      color = 0;
    } else
      color = xmlNodeListGetRawString(doc, cur->children->next->next->next->children, 0);
    user_id = find_user(user);
    printf("Adding post %d (f=%d, t=%d u=%d) (user=%s) (date=%ld)\n",
           post_id, forum_id, topic_id, user_id, user, time);
    add_post(forum_id, topic_id, post_id, user_id, time, first_post, user,
             color);
    if (test_mode) nb_posts_per_topic++;
    first_post = 0;
    cur = cur->next;
  }
  xmlFreeDoc(doc);

  /* go next page (if any) */
  /* we reuse posts_next_topics.xslt, change that if needed */
  doc = transform(d, "posts_next_topics.xslt");
  if (!doc) {
    printf("PANIC: %s:%d error with transform?\n", __FILE__, __LINE__);
    exit(1);
  }
  root = xmlDocGetRootElement(doc);
  if (!root) {
    printf("PANIC: no root element.\n");
    exit(1);
  }
  cur = root->children;
  if (cur && cur->children && cur->children->content) {
    nstart = 0;
    sscanf(cur->children->content, "%d", &nstart);
    if (nstart > start) {
      start = nstart;
      xmlFreeDoc(doc);
      goto more;
    }
  }
  xmlFreeDoc(doc);
}

void proceed_topics(void)
{
  int i;

  for (i=0; i<tl_size; i++) {
    if (test_mode) nb_posts_per_topic = 0;
    get_posts(tl+i);
  }
}

topic *find_topic_of_post(post *p)
{
  int i;
  if (!p->is_master) return 0;

  for (i=0; i<tl_size; i++)
    if (tl[i].topic_id == p->topic_id && tl[i].forum_id == p->forum_id)
      return tl+i;

  printf("PANIC: topic not found in %s\n", __FUNCTION__);
  exit(1);
}

void populate_topic_from_post(topic *t, post *p)
{
  t->topic_title = p->subject;
  t->topic_poster = p->poster;
  t->topic_time = p->time;
  t->topic_type = p->type;
  t->icon_id = p->icon;
  t->topic_locked = p->do_lock_topic;
  /* poll already populated */
  /* (yes, it's weird, but you know...) */
}

void do_post(post *p)
{
  char *d;
  xmlDocPtr doc;
  xmlNodePtr root;
  xmlNodePtr cur;
  int did_something = 0;
  topic *t = find_topic_of_post(p);

  printf("Proceeding post %d (f=%d, t=%d)\n", p->post_id, p->forum_id, p->topic_id);
  d = cache_get(EDITPOST, p->forum_id, p->topic_id, p->post_id);

  doc = transform(d, "posts_proceed_post.xslt");
  if (!doc) {
    printf("PANIC: %s:%d error with transform?\n", __FILE__, __LINE__);
    exit(1);
  }
  root = xmlDocGetRootElement(doc);
  if (!root) {
    printf("PANIC: no root element.\n");
    exit(1);
  }
  cur = root->children;
  NP(cur);
  while (cur) {
    did_something = 1;
    /* we may have several icon... let's take the bigger only */
    NP(cur->name);
    if (!strcmp(cur->name, "icon")) {
      int icon;
      NP(cur->children);
      NP(cur->children->content);
      printf("  found icon (%s)\n", cur->children->content);
      sscanf(cur->children->content, "%d", &icon);
      if (icon > p->icon) p->icon = icon;
    } else if (!strcmp(cur->name, "subject")) {
      NP(cur->children);
      p->subject = xmlNodeListGetRawString(doc, cur->children, 0);
      printf("  found subject (%s)\n", p->subject);
    } else if (!strcmp(cur->name, "content")) {
      NP(cur->children);
      p->content = xmlNodeListGetRawString(doc, cur->children, 0);
      printf("  found content (%s)\n", p->content);
    } else if (!strcmp(cur->name, "do_bbcode")) {
      NP(cur->children);
      NP(cur->children->content);
      printf("  found do_bbcode (%s)\n", cur->children->content);
      sscanf(cur->children->content, "%d", &p->do_bbcode);
    } else if (!strcmp(cur->name, "do_smilies")) {
      NP(cur->children);
      NP(cur->children->content);
      printf("  found do_smilies (%s)\n", cur->children->content);
      sscanf(cur->children->content, "%d", &p->do_smilies);
    } else if (!strcmp(cur->name, "do_magic_url")) {
      NP(cur->children);
      NP(cur->children->content);
      printf("  found do_magic_url (%s)\n", cur->children->content);
      sscanf(cur->children->content, "%d", &p->do_magic_url);
    } else if (!strcmp(cur->name, "do_attach_sig")) {
      NP(cur->children);
      NP(cur->children->content);
      printf("  found do_attach_sig (%s)\n", cur->children->content);
      sscanf(cur->children->content, "%d", &p->do_attach_sig);
    } else if (!strcmp(cur->name, "do_lock_topic")) {
      NP(cur->children);
      NP(cur->children->content);
      printf("  found do_lock_topic (%s)\n", cur->children->content);
      sscanf(cur->children->content, "%d", &p->do_lock_topic);
    } else if (!strcmp(cur->name, "do_lock_post")) {
      NP(cur->children);
      NP(cur->children->content);
      printf("  found do_lock_post (%s)\n", cur->children->content);
      sscanf(cur->children->content, "%d", &p->do_lock_post);
    } else if (!strcmp(cur->name, "type")) {
      NP(cur->children);
      NP(cur->children->content);
      printf("  found type (%s)\n", cur->children->content);
      sscanf(cur->children->content, "%d", &p->type);
    } else if (!strcmp(cur->name, "stick_topic_time_limit")) {
      NP(cur->children);
      NP(cur->children->content);
      printf("  found stick_topic_time_limit (%s)\n", cur->children->content);
      sscanf(cur->children->content, "%d", &p->stick_topic_time_limit);
    } else if (!strcmp(cur->name, "poll")) {
      /* NP sutff carefully chosen, some cases are allowed! (empty
       * question and empty options so cur->children->children
       * and cur->children->next->children are not checked)
       */
      NP(cur->children);
      NP(cur->children->next);
      NP(cur->children->next->next);
      NP(cur->children->next->next->children);
      NP(cur->children->next->next->children->content);
      NP(cur->children->next->next->next);
      NP(cur->children->next->next->next->children);
      NP(cur->children->next->next->next->children->content);
      NP(cur->children->next->next->next->next);
      NP(cur->children->next->next->next->next->children);
      NP(cur->children->next->next->next->next->children->content);
      if (!t) { printf("PANIC: impossible? %s\n", __FUNCTION__); exit(1); }
      printf("  found poll\n");
      t->has_poll = 1;
      t->poll_question = xmlNodeListGetRawString(doc,
          cur->children->children, 0);
      printf("    poll question (%s)\n", t->poll_question);
      if (!t->poll_question) t->poll_question = "";
      t->poll_options = xmlNodeListGetRawString(doc,
          cur->children->next->children, 0);
      printf("    poll options (%s)\n", t->poll_options);
      if (!t->poll_options) t->poll_options = "";
      printf("    options_per_user (%s)\n",
          cur->children->next->next->children->content);
      sscanf(cur->children->next->next->children->content, "%d",
             &t->options_per_user);
      printf("    poll_length (%s)\n",
          cur->children->next->next->next->children->content);
      sscanf(cur->children->next->next->next->children->content, "%d",
             &t->poll_length);
      printf("    poll_vote_change (%s)\n",
          cur->children->next->next->next->next->children->content);
      sscanf(cur->children->next->next->next->next->children->content, "%d",
             &t->poll_vote_change);
    } else
      printf("  UNHANDLED!! %s\n", cur->name);
    cur = cur->next;
  }
  xmlFreeDoc(doc);

  if (t) populate_topic_from_post(t, p);

  if (!did_something) { printf("PANIC: do_post did nothing...\n"); exit(1); }

  /* heck, last_post_subject must be set... */
  if (!t) t = find_topic(p->forum_id, p->topic_id);
  t->topic_last_post_subject = p->subject;
}

void proceed_posts(void)
{
  int i;

  for (i=0; i<pl_size; i++) do_post(pl+i);
  printf("Posts have been proceeded (%d posts found).\n", pl_size);
}

#define Ss do { fprintf(f, "fwrite($f, \""); } while (0) /**/
#define Se do { fprintf(f, "\");\n"); } while (0) /**/
#define Sfs do { fprintf(f, "fwrite($f, "); } while (0) /**/
#define Sfe do { fprintf(f, ");\n"); } while (0) /**/
#define S(x) do { fprintf(f, "%s", x); } while (0) /**/
#define I(x) do { fprintf(f, "%d", x); } while (0) /**/
#define LI(x) do { fprintf(f, "%ld", x); } while (0) /**/
void save_topics(void)
{
  int i;
  FILE *f = fopen("topics.php", "wb");
  if (!f) { printf("PANIC: topics.php %s\n", __FUNCTION__); exit(1); }

  fprintf(f, "<?php\n");

  for (i=0; i<tl_size; i++) {
    printf("Saving topic (topic=%d forum=%d)\n",
           tl[i].topic_id, tl[i].forum_id);
    printf("  title %p first poster name %p color %p last poster name %p color %p subject %p\n",
           tl[i].topic_title, tl[i].topic_first_poster_name,
           tl[i].topic_first_poster_colour, tl[i].topic_last_poster_name,
           tl[i].topic_last_poster_colour, tl[i].topic_last_post_subject);
    Ss;
    S("(");
    I(tl[i].topic_id); S(", ");
    I(tl[i].forum_id); S(", ");
    I(tl[i].icon_id); S(", '");
    Se;
    Sfs;
    S("mysql_real_escape_string(\"");
    S(sql_escape(tl[i].topic_title)); S("\")");
    Sfe; Ss; S("', ");
    I(tl[i].topic_poster); S(", ");
    LI(tl[i].topic_time); S(", ");
    /* ITEM_LOCKED = 1 and ITEM_UNLOCKED = 0
     * (we don't handle ITEM_MOVED, ask George Sodini) (ah, too late...)
     */
    if (tl[i].topic_locked) I(1); else I(0); S(", ");
    I(tl[i].topic_type); S(", ");
    I(tl[i].topic_replies); S(", ");
    I(tl[i].topic_replies); S(", ");
    I(tl[i].topic_first_post_id); S(", ");
    S("'"); Se; Sfs; S("mysql_real_escape_string(\"");
    S(sql_escape(tl[i].topic_first_poster_name)); S("\")"); Sfe; Ss; S("', ");
    S("'"); Se; Sfs; S("mysql_real_escape_string(\"");
    S(sql_escape(tl[i].topic_first_poster_colour)); S("\")"); Sfe; Ss; S("', ");
    I(tl[i].topic_last_post_id); S(", ");
    I(tl[i].topic_last_poster_id); S(", ");
    S("'"); Se; Sfs; S("mysql_real_escape_string(\"");
    S(sql_escape(tl[i].topic_last_poster_name)); S("\")"); Sfe; Ss; S("', ");
    S("'"); Se; Sfs; S("mysql_real_escape_string(\"");
    S(sql_escape(tl[i].topic_last_poster_colour)); S("\")"); Sfe; Ss; S("', ");
    S("'"); Se; Sfs; S("mysql_real_escape_string(\"");
    S(sql_escape(tl[i].topic_last_post_subject)); S("\")"); Sfe; Ss; S("', ");
    LI(tl[i].topic_last_post_time);
    /* we don't do polls */
    S(")");
    if (i != tl_size-1) S(", ");
    Se;
  }
  Ss; S(";\n"); Se;
  fprintf(f, "?>\n");
  fclose(f);
}

void save_posts(void)
{
  int i;
  FILE *f = fopen("posts.php", "wb");
  if (!f) { printf("PANIC: topics.php %s\n", __FUNCTION__); exit(1); }

  fprintf(f, "<?php\n");

  for (i=0; i<pl_size; i++) {
    Ss;
    S("(");
    I(pl[i].post_id); S(", ");
    I(pl[i].topic_id); S(", ");
    I(pl[i].forum_id); S(", ");
    I(pl[i].poster); S(", ");
    I(pl[i].icon); S(", ");
    LI(pl[i].time); S(", ");
    S(pl[i].do_bbcode ? "true" : "false"); S(", ");
    S(pl[i].do_smilies ? "true" : "false"); S(", ");
    S(pl[i].do_magic_url ? "true" : "false"); S(", ");
    S(pl[i].do_attach_sig ? "true" : "false"); S(", '");
    Se; Sfs; S("mysql_real_escape_string(\"");
    S(sql_escape(pl[i].subject)); S("\")"); Sfe; Ss; S("', '");

    Se;
    fprintf(f, "$m = new parse_message();\n");
    fprintf(f, "$m->message=\"%s\";\n", sql_escape(pl[i].content));
    fprintf(f, "$mmd5 = md5($m->message);\n");
    fprintf(f, "$m->parse(%s, %s, %s, true, true, true, true);\n",
            pl[i].do_bbcode ? "true" : "false",
            pl[i].do_magic_url ? "true" : "false",
            pl[i].do_smilies ? "true" : "false");
    fprintf(f, "fwrite($f, mysql_real_escape_string($m->message));\n");
    Ss; S("', '"); Se;
    fprintf(f, "fwrite($f, $m->bbcode_bitfield);\n");
    Ss; S("', '"); Se;
    fprintf(f, "fwrite($f, $m->bbcode_uid);\n");
    Ss; S("', "); S(pl[i].do_lock_post ? "true" : "false"); S(", '");
    Se; Sfs; S("$mmd5"); Sfe; Ss;
    S("')");
    if (i != pl_size-1) S(", ");
    Se;
  }
  Ss; S(";\n"); Se;
  fprintf(f, "?>\n");
  fclose(f);
}

forum *find_forum(int forum_id)
{
  int i;
  for (i=0; i<fl_size; i++)
    if (forum_list[i].forum_id == forum_id) return forum_list+i;
  printf("PANIC: %s\n", __FUNCTION__);
  exit(1);
}

void update_forums(void)
{
  forum *f;
  forum *p;
  int i;

  /* we browse topics and set forum's stuff */
  for (i=0; i < tl_size; i++) {
    f = find_forum(tl[i].forum_id);
    f->forum_posts += tl[i].topic_replies + 1;
    f->forum_topics++;
    if (tl[i].topic_last_post_time < f->forum_last_post_time) continue;
    f->forum_last_post_id = tl[i].topic_last_post_id;
    f->forum_last_poster_id = tl[i].topic_last_poster_id;
    f->forum_last_post_subject = tl[i].topic_last_post_subject;
    f->forum_last_post_time = tl[i].topic_last_post_time;
    f->forum_last_poster_name = tl[i].topic_last_poster_name;
    f->forum_last_poster_colour = tl[i].topic_last_poster_colour;
  }

  /* we add up data to the parent */
  /* running the list of forums from the end since we added parents first
   * and children next and we need to work children first then parents
   */
  for (i=fl_size-1; i>=0; i--) {
    f = forum_list + i;
    /* if no parent, do nothing */
    if (f->parent_id == -1) continue;
    p = find_forum(f->parent_id);
    /* counting up? not counting up? */
    //p->forum_posts += f->forum_posts;
    //p->forum_topics += f->forum_topics;
    if (p->forum_last_post_time > f->forum_last_post_time) continue;
    p->forum_last_post_id = f->forum_last_post_id;
    p->forum_last_poster_id = f->forum_last_poster_id;
    p->forum_last_post_subject = f->forum_last_post_subject;
    p->forum_last_post_time = f->forum_last_post_time;
    p->forum_last_poster_name = f->forum_last_poster_name;
    p->forum_last_poster_colour = f->forum_last_poster_colour;
  }
}

void save_forums(void)
{
  int i;
  forum *c;
  FILE *f = fopen("update_forums.php", "wb");
  if (!f) { printf("PANIC: update_forums.php %s\n", __FUNCTION__); exit(1); }

  fprintf(f, "<?php\n");

  for (i=0; i<fl_size; i++) {
    c = forum_list + i;
    Ss;
    S("UPDATE phpbb_forums SET ");
    S("forum_posts = "); I(c->forum_posts);
    S(", forum_topics = "); I(c->forum_topics);
    S(", forum_topics_real = "); I(c->forum_topics);
    S(", forum_last_post_id = "); I(c->forum_last_post_id);
    S(", forum_last_poster_id = "); I(c->forum_last_poster_id);
    S(", forum_last_post_subject = ");
    S("'"); Se; Sfs; S("mysql_real_escape_string(\"");
    S(sql_escape(c->forum_last_post_subject)); S("\")"); Sfe; Ss; S("'");
    S(", forum_last_post_time = "); LI(c->forum_last_post_time);
    S(", forum_last_poster_name = ");
    S("'"); Se; Sfs; S("mysql_real_escape_string(\"");
    S(sql_escape(c->forum_last_poster_name)); S("\")"); Sfe; Ss; S("'");
    S(", forum_last_poster_colour = ");
    S("'"); Se; Sfs; S("mysql_real_escape_string(\"");
    S(sql_escape(c->forum_last_poster_colour)); S("\")"); Sfe; Ss; S("'");
    S(" WHERE forum_id = ");
    I (c->forum_id);
    S(";\n");
    Se;
  }

  fprintf(f, "?>\n");
  fclose(f);
}

int main(int n, char **v)
{
  read_users();

  com_init("posts", n, v);

while (0) {
  char *d = cache_get(VIEWFORUM, 37, 0, 0);
  xmlDocPtr doc = transform(d, "posts_get_topics.xslt");
  xmlNodePtr root;
  if (!doc) exit(1);
  root = xmlDocGetRootElement(doc);
  xmlFreeDoc(doc);
}

  login();
  acp_login();
  get_forums();
  proceed_topics();
  proceed_posts();
  update_forums();

  save_topics();
  save_posts();
  save_forums();

  logout();

  com_fini();

  return 0;
}
