#include "buffer.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

char *prog;
char *input;
char output[1024];
char *pattern;
int pindex;

FILE *fi;
FILE *fo;

do_buf head;
do_buf post;

int maxsize = 2000000;
int size;

void read_header(void)
{
  int c;
  while (1) {
    c = fgetc(fi);
    if (c == EOF) goto err;
    if (c == ')') break;
    add_buffer_char(&head, c);
    if (c == '\n') {
      fprintf(fo, "%s", head.data);
      reset_buffer(&head);
    }
  }

  while (c != '(') {
    add_buffer_char(&head, c);
    c = fgetc(fi);
    if (c == EOF) goto err;
  }
  ungetc(c, fi);
  add_buffer_char(&head, 0);
  return;

err:
  printf("bad file: %s\n", input);
  exit(1);
}

void read_string(void)
{
  int c;

  /* read up (and including) to ', take care of \' */
  while (1) {
    c = fgetc(fi);
    if (c == EOF) goto err;
    add_buffer_char(&post, c);
    if (c == '\'') break;
    if (c != '\\') continue;
    c = fgetc(fi);
    if (c == EOF) goto err;
    add_buffer_char(&post, c);
  }

  return;

err:
  printf("bad file: %s\n", input);
  exit(1);
}

/* return 0 if more post, 1 if last post */
int read_post(void)
{
  int c;

  reset_buffer(&post);

  while (1) {
    c = fgetc(fi);
    if (c == EOF) goto err;
    add_buffer_char(&post, c);
    if (c == '\'') read_string();
    if (c == ')') break;
  }

  add_buffer_char(&post, 0);

  /* get separator */
  while (1) {
    c = fgetc(fi);
    if (c == EOF) goto err;
    if (c == ',') break;
    if (c == ';') return 1;
    if (!isspace(c)) goto err;
  }

  /* we've got more */
  while (1) {
    c = fgetc(fi);
    if (c == EOF) goto err;
    if (c == '(') break;
  }

  ungetc(c, fi);
  return 0;

err:
  printf("bad file: %s\n", input);
  exit(1);
}

int nb_post;

void add_post(void)
{
  static int first_post = 1;

  nb_post++;

  if (size + post.size > maxsize-2) {
    fprintf(fo, ";\n");
    fclose(fo);
    pindex++;
    sprintf(output, pattern, pindex);
    fo = fopen(output, "wb");
    if (!fo) { perror(""); printf("error opening %s\n", output); exit(1); }
    fprintf(fo, "%s", head.data);
    size = head.size;
    first_post = 1;
  }
  if (!first_post) { fprintf(fo, ", "); size += 2; }
  fprintf(fo, "%s", post.data);
  size += post.size;
  first_post = 0;
}

void go(void)
{
  fi = fopen(input, "rb");
  if (!fi) { perror(""); printf("error opening %s\n", input); exit(1); }

  sprintf(output, pattern, pindex);
  fo = fopen(output, "wb");
  if (!fo) { perror(""); printf("error opening %s\n", output); exit(1); }

  read_header();

  fprintf(fo, "%s", head.data);
  size = head.size;

  while (!read_post()) add_post();
  add_post();
  
  printf("Proceeded %d posts.\n", nb_post);

  fclose(fi);
  fprintf(fo, ";\n");
  fclose(fo);
}

void verify_pattern(void)
{
  int gotit = 0;
  char *c = pattern;
  /* one and only one %d allowed, no other %X allowed */
  while (*c) {
    if (*c != '%') { c++; continue; }
    if (gotit) goto error;
    gotit = 1;
    c++;
    if (*c != 'd') goto error;
    c++;
  }
  if (!gotit) goto error;

  return;

error:
  printf("malformed pattern: %s\n", pattern);
  exit(1);
}

void usage(void)
{
  printf("usage: %s <options> <input file> <pattern>\n", prog);
  printf("options:\n");
  printf("    -s <size>: maximum size of file in bytes\n");
  printf("pattern: something like 'posts_%%d.sql'\n");
  printf("    %%d must appear somewhere, it will be used to save\n");
  printf("    the several sql files.\n");
  exit(0);
}

int main(int n, char **v)
{
  int i = 1;
  prog = v[0];
  while (i != n) {
    if (!strcmp(v[i], "-s")) {
      i++; if (i == n) usage();
      sscanf(v[i], "%d", &maxsize);
      i++; continue;
    }
    /* no known option, must be input */
    if (!input) {
      input = v[i];
      i++; continue;
    }
    /* input already set, must be pattern */
    if (!pattern) {
      pattern = v[i];
      i++; continue;
    }
    /* pattern already set, error */
    usage();
  }
  if (!input || !pattern) usage();

  verify_pattern();

  if (strlen(pattern) + 20 > 1023) {
    printf("Pattern too long.\n");
    exit(1);
  }

  go();

  return 0;
}
