#include "xml.h"

#include <libxslt/xslt.h>
#include <libxslt/transform.h>
#include <libxml/HTMLparser.h>

void init_xml(void)
{
  xmlInitMemory();
  LIBXML_TEST_VERSION
}

void close_xml(void)
{
  xsltCleanupGlobals();
  xmlCleanupParser();
  xmlMemoryDump();
}

/* return 0 if error, a pointer to a xmlDoc if ok
 * h: the http data received (with headers)
 * x: the xslt filename to use to transform the data
 */
void *transform(char *h, char *x)
{
  xsltStylesheetPtr cur = 0;
  xmlDocPtr doc = 0;
  xmlDocPtr res = 0;
  const char *params[1];

  cur = xsltParseStylesheetFile((const xmlChar *)x);
  if (!cur) {
    fprintf(stderr, "transform error, can't read %s\n", x);
    goto error;
  }

  doc = htmlParseDoc(h, 0);
  if (!doc) goto error;
  params[0] = 0;
  res = xsltApplyStylesheet(cur, doc, params);
  if (!res) goto error;

  xmlFreeDoc(doc);
  xsltFreeStylesheet(cur);

  return res;

error:
  if (res) xmlFreeDoc(res);
  if (doc) xmlFreeDoc(doc);
  if (cur) xsltFreeStylesheet(cur);
  return 0;
}

void free_doc(void *doc)
{
  xmlFreeDoc(doc);
}
